
// depmain.cpp
// Copyright 2022 Matthew Rickard
// This file is part of dep

// dep is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "precomp.h"
#include "logMsg.h"
#include <fstream>
#include "dep.h"
#include <string.h>
#include "MacroNames.h"
#include <sys/stat.h>
#include "AbstractProcess.h"
#include "matches.h"
#include "DepTree.h"
#include "argv.h"

int atoidefault(const char *s, int def) {
  if (isdigit(*s))
    return atoi(s);
  return def;
}

const char *bs = "dep version 0.51 copyright 2022 Matthew Rickard. Buildstamp " __DATE__ " " __TIME__;

int main(int argc, const char *argv[]) {
  olib::argc = argc;
  olib::argv = (char **)argv;
  Cmdspec cmdspec("-f: -h-help--help -V--version -r--remake -a--remakeall -d--dryrun\n"
  " -q--quiet -j: --treeincludes --ruleincludes\n"
  "--noincludes --bothincludes --maxnonincludes: --sense=newer,rule,modtime,size\n"
  "--noreadcache --nowritecache --writecache --confirmchanges --mkdir --noskip\n"
  "--tracefile: --fixedorder -v?",
    0, Position(__FILE__, __LINE__));
  AutoCmdLine cmdline(&cmdspec);
  cmdline.processArgv(FRAGMENTTYPE_COMMAND, argc, argv);
  if (cmdline.haltline) {
    if (*argv[cmdline.paramIdx] == '-' || *argv[cmdline.paramIdx] == '/') {
      std::cerr << "dep: Unrecognized option " << argv[cmdline.paramIdx] << std::endl
        << "Exiting" << std::endl;
      return 1;
    }
  }
  if (cmdline.optionMap.count("-h")) {
    std::cout << "Usage: dep " << cmdspec.spec << std::endl;
    return 0;
  }
  if (cmdline.optionMap.count("-V")) {
    std::cout << "dep version 0.51" << std::endl;
    return 0;
  }
  if (!cmdline.optionMap.count("-f"))
    cmdline.optionMap["-f"].push_back("DepFile");

  RuleSet ruleset;
  ruleset.cmdline = &cmdline;

  File out("stdout", 1, std::cout);
  int readCache = 1, writeCache = 1;
  ruleset.outputFile = &out;
  if (cmdline.optionMap.count("-v")) {
    for (auto i = cmdline.optionMap["-v"].begin(); i != cmdline.optionMap["-v"].end(); i++)
      ruleset.interpretVerbosity(*i);
  }
  if (cmdline.optionMap.count("-q")) {
    ruleset.interpretVerbosity("-1");
  }
  if (cmdline.optionMap.count("--confirmchanges"))
    ruleset.confirmRuleChanges = 1;

  ruleset.optionTreeIncludes = 1; // This should be the fastest but has problems with different -I FIXED
                                  // And fails on full tests on alma due to mingw64 not installed FIXED
  ruleset.optionRuleIncludes = 0; // This should be the most reliable but has bugs FIXED
                                  // And fails check-acceptance

  if (cmdline.optionMap.count("-r"))
    ruleset.optionRebuildMarked = 1;
  if (cmdline.optionMap.count("-a"))
    ruleset.optionRebuildAll = 1;
  if (cmdline.optionMap.count("-d"))
    ruleset.optionDryrun = 1;
  if (cmdline.optionMap.count("--noskip")) {
    ruleset.optionNoSkip = 1;
  }
  if (cmdline.optionMap.count("--mkdir")) {
    ruleset.optionMkdir = 1;
  }
  if (cmdline.optionMap.count("--fixedorder")) {
    ruleset.optionFixedOrder = 1;
    ruleset.maxPipelines = 1; // this helps run things in fixed order for check-acceptance.
  }
  if (cmdline.optionMap.count("--noincludes")) {
    ruleset.optionTreeIncludes = 0;
    ruleset.optionRuleIncludes = 0;
  }
  else if (cmdline.optionMap.count("--treeincludes")) {
    ruleset.optionTreeIncludes = 1;
    ruleset.optionRuleIncludes = 0;
  }
  else if (cmdline.optionMap.count("--ruleincludes")) {
    ruleset.optionTreeIncludes = 0;
    ruleset.optionRuleIncludes = 1;
  }
  else if (cmdline.optionMap.count("--bothincludes")) {
    ruleset.optionTreeIncludes = 1;
    ruleset.optionRuleIncludes = 1;
  }
  if (cmdline.optionMap.count("--noreadcache")) {
    readCache = 0;
  }
  if (cmdline.optionMap.count("--nowritecache")) {
    writeCache = 0;
  }
  if (cmdline.optionMap.count("--writecache")) {
    writeCache = 2;
  }
  if (cmdline.optionMap.count("--tracefile")) {
    //ruleset.tracefile = cmdline.optionMap["--tracefile"].back();
    for (unsigned i = 0; i < cmdline.optionMap["--tracefile"].size(); i++)
    ruleset.tracefileSet.insert(cmdline.optionMap["--tracefile"][i]);
  }
  ruleset.senseNewer   = 1;
  ruleset.senseRule    = 1;
  ruleset.senseModTime = 0;
  ruleset.senseSize    = 0;
  //logMsg << "Checkpoint" << std::endl;
  //logMsg << "Checkpoint 1" << std::endl;
  if (cmdline.optionMap.count("--sense=")) {
    //logMsg << "Checkpoint 2" << std::endl;
    std::string sense = cmdline.optionMap["--sense="].back();
    if (sense.find("all") != std::string::npos) {
      ruleset.senseModTime = 1;
      ruleset.senseSize    = 1;
    }
    else {
      ruleset.senseNewer   = (sense.find("newer")   != std::string::npos);
      ruleset.senseRule    = (sense.find("rule")    != std::string::npos);
      ruleset.senseModTime = (sense.find("modtime") != std::string::npos);
      ruleset.senseSize    = (sense.find("size")    != std::string::npos);
    }
    //logMsg << logValue(sense) << std::endl;
    //logMsg << logValue(ruleset.senseNewer)   << std::endl;
    //logMsg << logValue(ruleset.senseRule)    << std::endl;
    //logMsg << logValue(ruleset.senseModTime) << std::endl;
    //logMsg << logValue(ruleset.senseSize)    << std::endl;
  }
  //else
    //logMsg << "Not doing --sense" << std::endl;
  if (cmdline.optionMap.count("-j"))
    ruleset.maxPipelines = atoidefault(cmdline.optionMap["-j"].back().c_str(), 1);
  if (cmdline.optionMap.count("--maxnonincludes"))
    ruleset.maxNonIncludes = atoidefault(cmdline.optionMap["--maxnonincludes"].back().c_str(), 1000);

  std::string filename = cmdline.optionMap["-f"].back().c_str();

  DepTree cacheTree;
  RuleSet cacheset;
  //logMsg << "Clearing verbosity in cacheset" << std::endl;
  cacheset.setVerbosity(-1);
  cacheset.tracefileSet = ruleset.tracefileSet;
  std::string cachefilename("." + filename + "Cache");

  Timer timer;

  // Read the cache first. It is faster and easier to read, and has some information that the
  // main tree can use.

#define SHOW_TIMINGS 0

  if (readCache) {
    cacheset.outputFile = &out;
    cacheTree.ruleset = &cacheset;
    if (cacheTree.readFile(cachefilename)) {
      std::cerr << "(dep: Can't read cache file " << cachefilename << ")\n";
      readCache = 0;
    }
#if SHOW_TIMINGS
    logMsg << (timer.getMicroseconds(TIMER_RESET)) << " micros reading the caches" << std::endl;
#endif
  }

  DepTree depTree;
  depTree.ruleset = &ruleset;
  ruleset.cacheset = readCache? &cacheset: 0;
  int result = depTree.readFile(filename);
#if SHOW_TIMINGS
  logMsg << (timer.getMicroseconds(TIMER_RESET)) << " micros reading the DepFiles" << std::endl;
#endif
  if (result != 0) {
    std::cerr << "dep: Can't read " << cmdline.optionMap["-f"].back() << std::endl;
    return 1;
  }

  File outputFile("stdout", 1, std::cout);
  result = ruleset.processCmd(argc - cmdline.paramIdx, argv + cmdline.paramIdx, &outputFile, &depTree);
#if SHOW_TIMINGS
  logMsg << (timer.getMicroseconds(TIMER_RESET)) << " micros processing the command" << std::endl;
#endif
  if (writeCache && ruleset.startedSomething || writeCache == 2) {
    //logMsg << logValue(writeCache) << ' ' << logValue(ruleset.startedSomething) << std::endl;
    ruleset.combineWithCache(cacheset);
    copyFile(cachefilename, cachefilename + ".bak");
    std::ofstream ofs(cachefilename, std::ios::out | std::ios::trunc);
    //logMsg << "Outputting to " << logValue(cachefilename) << std::endl;
    ruleset.outputTo(ofs);
    //logMsg << "Output to " << logValue(cachefilename) << std::endl;
  }
#if SHOW_TIMINGS
  logMsg << (timer.getMicroseconds(TIMER_RESET)) << " micros writing the cache" << std::endl;
#endif

  if (result == -1)
    return 1;
  if (WIFEXITED(result)) {
    if (WEXITSTATUS(result)) {
      //logMsg << "Checkpoint" << std::endl;
      std::cerr << "dep: child process exited with value " << WEXITSTATUS(result) << std::endl;
    }
    if (ruleset.failed)
      return 1;
    //return WEXITSTATUS(result);
  }
  if (WIFSIGNALED(result)) {
      //logMsg << "Checkpoint" << std::endl;
    std::cerr << "dep: child process terminated by signal "
      << WTERMSIG(result) << ' ' << SignalToMacroName(WTERMSIG(result)) << std::endl;
    return 13;
  }
  if (result == 65535) {
    //logMsg << "Checkpoint" << std::endl;
    std::cerr << "dep: couldn't start build command" << std::endl;
    return 133;
  }
  if (!WIFEXITED(result)) {
    logMsg << "Process terminated by something other than exit or signal "
      << logValue(result) << std::endl;
    return 113;
  }
  return 0;
}

